# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


from datetime import datetime, timedelta
from functools import reduce
from operator import add, and_, mul, or_, sub, truediv
from typing import Dict, List, Optional, Union

from django.contrib.gis.gdal import SpatialReference
from django.contrib.gis.geos import Polygon
from django.contrib.gis.measure import D
from django.db.models import F, Q, Value
from django.db.models.expressions import Expression

ArithmeticType = Union[Expression, F, Value, int, float]

# ------------------------------------------------------------------------------
# Filters
# ------------------------------------------------------------------------------


def combine(sub_filters: List[Q], combinator: str = "AND") -> Q:
    """Combine filters using a logical combinator"""
    op = and_ if combinator == "AND" else or_
    return reduce(lambda acc, q: op(acc, q) if acc else q, sub_filters)


def negate(sub_filter: Q) -> Q:
    """Negate a filter, opposing its meaning."""
    return ~sub_filter


OP_TO_COMP = {"<": "lt", "<=": "lte", ">": "gt", ">=": "gte", "<>": None, "=": "exact"}

INVERT_COMP: Dict[Optional[str], str] = {
    "lt": "gt",
    "lte": "gte",
    "gt": "lt",
    "gte": "lte",
}


def compare(
    lhs: Union[F, Value],
    rhs: Union[F, Value],
    op: str,
    mapping_choices: Optional[Dict[str, Dict[str, str]]] = None,
) -> Q:
    """Compare a filter with an expression using a comparison operation

    :param lhs: the field to compare
    :type lhs: :class:`django.db.models.F`
    :param rhs: the filter expression
    :type rhs: :class:`django.db.models.F`
    :param op: a string denoting the operation. one of ``"<"``, ``"<="``,
               ``">"``, ``">="``, ``"<>"``, ``"="``
    :type op: str
    :param mapping_choices: a dict to lookup potential choices for a
                            certain field.
    :type mapping_choices: dict[str, str]
    :return: a comparison expression object
    :rtype: :class:`django.db.models.Q`
    """
    comp = OP_TO_COMP[op]

    # if the left hand side is not a field reference, the comparison
    # can be be inverted to try if the right hand side is a field
    # reference.
    if not isinstance(lhs, F):
        lhs, rhs = rhs, lhs
        comp = INVERT_COMP.get(comp, comp)

    # if neither lhs and rhs are fields, we have to fail here
    if not isinstance(lhs, F):
        raise ValueError(f"Unable to compare non-field {lhs}")

    field_name = lhs.name

    if mapping_choices and field_name in mapping_choices:
        try:
            if isinstance(rhs, str):
                rhs = mapping_choices[field_name][rhs]
            elif hasattr(rhs, "value"):
                rhs = Value(mapping_choices[field_name][rhs.value])

        except KeyError as e:
            raise AssertionError("Invalid field value %s" % e)

    if comp:
        return Q(**{"%s__%s" % (lhs.name, comp): rhs})
    return ~Q(**{field_name: rhs})


def between(
    lhs: F, low: Union[F, Value], high: Union[F, Value], not_: bool = False
) -> Q:
    """Create a filter to match elements that have a value within a certain
    range.

    :param lhs: the field to compare
    :type lhs: :class:`django.db.models.F`
    :param low: the lower value of the range
    :type low:
    :param high: the upper value of the range
    :type high:
    :param not_: whether the range shall be inclusive (the default) or
                 exclusive
    :type not_: bool
    :return: a comparison expression object
    :rtype: :class:`django.db.models.Q`
    """
    q = Q(**{"%s__range" % lhs.name: (low, high)})
    return ~q if not_ else q


def like(
    lhs: F,
    pattern: str,
    nocase: bool = False,
    not_: bool = False,
    mapping_choices: Optional[Dict[str, Dict[str, str]]] = None,
) -> Q:
    """Create a filter to filter elements according to a string attribute
    using wildcard expressions.

    :param lhs: the field to compare
    :type lhs: :class:`django.db.models.F`
    :param rhs: the wildcard pattern: a string containing any number of '%'
                characters as wildcards.
    :type rhs: str
    :param case: whether the lookup shall be done case sensitively or not
    :type case: bool
    :param not_: whether the range shall be inclusive (the default) or
                 exclusive
    :type not_: bool
    :param mapping_choices: a dict to lookup potential choices for a
                            certain field.
    :type mapping_choices: dict[str, str]
    :return: a comparison expression object
    :rtype: :class:`django.db.models.Q`
    """
    parts = pattern.split("%")
    length = len(parts)

    if mapping_choices and lhs.name in mapping_choices:
        # special case when choices are given for the field:
        # compare statically and use 'in' operator to check if contained
        cmp_av = [
            (a, a.lower() if nocase else a) for a in mapping_choices[lhs.name].keys()
        ]

        for idx, part in enumerate(parts):
            if not part:
                continue

            cmp_p = part.lower() if nocase else part

            if idx == 0 and length > 1:  # startswith
                cmp_av = [a for a in cmp_av if a[1].startswith(cmp_p)]
            elif idx == 0:  # exact matching
                cmp_av = [a for a in cmp_av if a[1] == cmp_p]
            elif idx == length - 1:  # endswith
                cmp_av = [a for a in cmp_av if a[1].endswith(cmp_p)]
            else:  # middle
                cmp_av = [a for a in cmp_av if cmp_p in a[1]]

        q = Q(
            **{"%s__in" % lhs.name: [mapping_choices[lhs.name][a[0]] for a in cmp_av]}
        )

    else:
        i = "i" if nocase else ""
        q = None

        for idx, part in enumerate(parts):
            if not part:
                continue

            if idx == 0 and length > 1:  # startswith
                new_q = Q(**{"%s__%s" % (lhs.name, "%sstartswith" % i): part})
            elif idx == 0:  # exact matching
                new_q = Q(**{"%s__%s" % (lhs.name, "%sexact" % i): part})
            elif idx == length - 1:  # endswith
                new_q = Q(**{"%s__%s" % (lhs.name, "%sendswith" % i): part})
            else:  # middle
                new_q = Q(**{"%s__%s" % (lhs.name, "%scontains" % i): part})

            q = q & new_q if q else new_q

    return ~q if not_ else q


def contains(
    lhs: F,
    items: List[Union[F, Value]],
    not_: bool = False,
    mapping_choices: Optional[Dict[str, Dict[str, str]]] = None,
) -> Q:
    """Create a filter to match elements attribute to be in a list of choices.

    :param lhs: the field to compare
    :type lhs: :class:`django.db.models.F`
    :param items: a list of choices
    :type items: list
    :param not_: whether the range shall be inclusive (the default) or
                 exclusive
    :type not_: bool
    :param mapping_choices: a dict to lookup potential choices for a
                            certain field.
    :type mapping_choices: dict[str, str]
    :return: a comparison expression object
    :rtype: :class:`django.db.models.Q`
    """

    if mapping_choices is not None and lhs.name in mapping_choices:

        def map_value(
            item: Union[str, Value], choices: Dict[str, str]
        ) -> Union[str, Value]:
            try:
                if isinstance(item, str):
                    item = choices[item]
                elif isinstance(item, Value):
                    item = Value(choices[item.value])

            except KeyError as e:
                raise AssertionError("Invalid field value %s" % e)
            return item

        items = [map_value(item, mapping_choices[lhs.name]) for item in items]

    q = Q(**{"%s__in" % lhs.name: items})
    return ~q if not_ else q


def null(lhs: F, not_: bool = False) -> Q:
    """Create a filter to match elements whose attribute is (not) null

    :param lhs: the field to compare
    :type lhs: :class:`django.db.models.F`
    :param not_: whether the range shall be inclusive (the default) or
                 exclusive
    :type not_: bool
    :return: a comparison expression object
    :rtype: :class:`django.db.models.Q`
    """
    return Q(**{"%s__isnull" % lhs.name: not not_})


def temporal(lhs: F, time_or_period: Value, op: str) -> Q:
    """Create a temporal filter for the given temporal attribute.

    :param lhs: the field to compare
    :type lhs: :class:`django.db.models.F`
    :param time_or_period: the time instant or time span to use as a filter
    :type time_or_period: :class:`datetime.datetime` or a tuple of two
                          datetimes or a tuple of one datetime and one
                          :class:`datetime.timedelta`
    :param op: the comparison operation. one of ``"BEFORE"``,
               ``"BEFORE OR DURING"``, ``"DURING"``, ``"DURING OR AFTER"``,
               ``"AFTER"``.
    :type op: str
    :return: a comparison expression object
    :rtype: :class:`django.db.models.Q`
    """
    assert op in ("BEFORE", "BEFORE OR DURING", "DURING", "DURING OR AFTER", "AFTER")
    time_or_period = time_or_period.value
    low: Union[datetime, timedelta, None] = None
    high: Union[datetime, timedelta, None] = None
    if op in ("BEFORE", "AFTER"):
        assert isinstance(time_or_period, datetime)
        if op == "BEFORE":
            high = time_or_period
        else:
            low = time_or_period
    else:
        low, high = time_or_period
        low = low.value if isinstance(low, Value) else low
        high = high.value if isinstance(high, Value) else high
        assert isinstance(low, datetime) or isinstance(high, datetime)

        if isinstance(low, timedelta) and isinstance(high, datetime):
            low = high - low
        if isinstance(low, datetime) and isinstance(high, timedelta):
            high = low + high

    if low and high:
        return Q(**{"%s__range" % lhs.name: (low, high)})
    elif low:
        return Q(**{"%s__gte" % lhs.name: low})
    else:
        return Q(**{"%s__lte" % lhs.name: high})


def time_interval(
    time_or_period: Value,
    containment: str = "overlaps",
    begin_time_field: str = "begin_time",
    end_time_field: str = "end_time",
) -> Q:
    """ """

    gt_op = "__gte"
    lt_op = "__lte"

    is_slice = len(time_or_period) == 1
    if len(time_or_period) == 1:
        is_slice = True
        value = time_or_period[0]
    else:
        is_slice = False
        low, high = time_or_period

    if is_slice or (high == low and containment == "overlaps"):
        return Q(
            **{
                begin_time_field + "__lte": time_or_period[0],
                end_time_field + "__gte": time_or_period[0],
            }
        )

    elif high == low:
        return Q(**{begin_time_field + "__gte": value, end_time_field + "__lte": value})

    else:
        q = Q()
        # check if the temporal bounds must be strictly contained
        if containment == "contains":
            if high is not None:
                q &= Q(**{end_time_field + lt_op: high})
            if low is not None:
                q &= Q(**{begin_time_field + gt_op: low})
        # or just overlapping
        else:
            if high is not None:
                q &= Q(**{begin_time_field + lt_op: high})
            if low is not None:
                q &= Q(**{end_time_field + gt_op: low})
        return q


UNITS_LOOKUP = {"kilometers": "km", "meters": "m"}


INVERT_SPATIAL_OP = {
    "WITHIN": "CONTAINS",
    "CONTAINS": "WITHIN",
}


def spatial(
    lhs: Union[F, Value],
    rhs: Union[F, Value],
    op: str,
    pattern: Optional[str] = None,
    distance: Optional[float] = None,
    units: Optional[str] = None,
) -> Q:
    """Create a spatial filter for the given spatial attribute.

    :param lhs: the field to compare
    :type lhs: :class:`django.db.models.F`
    :param rhs: the time instant or time span to use as a filter
    :type rhs:
    :param op: the comparison operation. one of ``"INTERSECTS"``,
               ``"DISJOINT"``, `"CONTAINS"``, ``"WITHIN"``,
               ``"TOUCHES"``, ``"CROSSES"``, ``"OVERLAPS"``,
               ``"EQUALS"``, ``"RELATE"``, ``"DWITHIN"``, ``"BEYOND"``
    :type op: str
    :param pattern: the spatial relation pattern
    :type pattern: str
    :param distance: the distance value for distance based lookups:
                     ``"DWITHIN"`` and ``"BEYOND"``
    :type distance: float
    :param units: the units the distance is expressed in
    :type units: str
    :return: a comparison expression object
    :rtype: :class:`django.db.models.Q`
    """

    assert op in (
        "INTERSECTS",
        "DISJOINT",
        "CONTAINS",
        "WITHIN",
        "TOUCHES",
        "CROSSES",
        "OVERLAPS",
        "EQUALS",
        "RELATE",
        "DWITHIN",
        "BEYOND",
    )

    # if the left hand side is not a field reference, the comparison
    # can be be inverted to try if the right hand side is a field
    # reference.
    if not isinstance(lhs, F):
        lhs, rhs = rhs, lhs
        op = INVERT_SPATIAL_OP.get(op, op)

    # if neither lhs and rhs are fields, we have to fail here
    if not isinstance(lhs, F):
        raise ValueError(f"Unable to compare non-field {lhs}")

    return Q(**{"%s__%s" % (lhs.name, op.lower()): rhs})


def spatial_relate(lhs: Union[F, Value], rhs: Union[F, Value], pattern: str) -> Q:

    if not isinstance(lhs, F):
        # TODO: cannot yet invert pattern -> raise
        raise ValueError(f"Unable to compare non-field {lhs}")

    return Q(**{"%s__relate" % lhs.name: (rhs, pattern)})


def spatial_distance(
    lhs: Union[F, Value], rhs: Union[F, Value], op: str, distance: float, units: str
) -> Q:
    if not isinstance(lhs, F):
        lhs, rhs = rhs, lhs

    # if neither lhs and rhs are fields, we have to fail here
    if not isinstance(lhs, F):
        raise ValueError(f"Unable to compare non-field {lhs}")

    # TODO: maybe use D.unit_attname(units)
    d = D(**{UNITS_LOOKUP[units]: distance})
    if op == "DWITHIN":
        return Q(**{"%s__distance_lte" % lhs.name: (rhs, d, "spheroid")})
    return Q(**{"%s__distance_gte" % lhs.name: (rhs, d, "spheroid")})


def bbox(
    lhs: F,
    minx: float,
    miny: float,
    maxx,
    maxy: float,
    crs: Optional[str] = None,
    bboverlaps: bool = True,
) -> Q:
    """Create a bounding box filter for the given spatial attribute.

    :param lhs: the field to compare
    :param minx: the lower x part of the bbox
    :type minx: float
    :param miny: the lower y part of the bbox
    :type miny: float
    :param maxx: the upper x part of the bbox
    :type maxx: float
    :param maxy: the upper y part of the bbox
    :type maxy: float
    :param crs: the CRS the bbox is expressed in
    :type crs: str
    :type lhs: :class:`django.db.models.F`
    :return: a comparison expression object
    :rtype: :class:`django.db.models.Q`
    """
    box = Polygon.from_bbox((minx, miny, maxx, maxy))

    if crs:
        box.srid = SpatialReference(crs).srid
        box.transform(4326)

    if bboverlaps:
        return Q(**{"%s__bboverlaps" % lhs.name: box})
    return Q(**{"%s__intersects" % lhs.name: box})


def attribute(name: str, field_mapping: Optional[Dict[str, str]] = None) -> F:
    """Create an attribute lookup expression using a field mapping dictionary.

    :param name: the field filter name
    :type name: str
    :param field_mapping: the dictionary to use as a lookup.
    :rtype: :class:`django.db.models.F`
    """
    if field_mapping:
        field = field_mapping.get(name, name)
    else:
        field = name
    return F(field)


def literal(value) -> Value:
    return Value(value)


OP_TO_FUNC = {"+": add, "-": sub, "*": mul, "/": truediv}


def arithmetic(lhs: ArithmeticType, rhs: ArithmeticType, op: str) -> ArithmeticType:
    """Create an arithmetic filter

    :param lhs: left hand side of the arithmetic expression. either a
                scalar or a field lookup or another type of expression
    :param rhs: same as `lhs`
    :param op: the arithmetic operation. one of ``"+"``, ``"-"``, ``"*"``,
               ``"/"``
    :rtype: :class:`django.db.models.F`
    """
    func = OP_TO_FUNC[op]
    return func(lhs, rhs)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from datetime import date, datetime, time, timedelta, timezone
from typing import Any, Callable, Dict, Optional, Tuple, Union

import shapely.geometry

from ... import ast, values
from ...util import like_pattern_to_re, parse_datetime
from ..evaluator import Evaluator, handle

COMPARISON_MAP = {
    ast.ComparisonOp.EQ: "==",
    ast.ComparisonOp.NE: "!=",
    ast.ComparisonOp.LT: "<",
    ast.ComparisonOp.LE: "<=",
    ast.ComparisonOp.GT: ">",
    ast.ComparisonOp.GE: ">=",
}

ARITHMETIC_MAP = {
    ast.ArithmeticOp.ADD: "+",
    ast.ArithmeticOp.SUB: "-",
    ast.ArithmeticOp.MUL: "*",
    ast.ArithmeticOp.DIV: "/",
}

ARRAY_COMPARISON_OP_MAP = {
    ast.ArrayComparisonOp.AEQUALS: "==",
    ast.ArrayComparisonOp.ACONTAINS: ">=",
    ast.ArrayComparisonOp.ACONTAINEDBY: "<=",
    ast.ArrayComparisonOp.AOVERLAPS: "&",
}


class NativeEvaluator(Evaluator):
    """This evaluator type allows to create a filter that can be used to
    filter objects or dicts.

    The filter is built using Python expressions which are then parsed
    using eval. The result is a callable object that can be used in any
    circumstance a normal function would.
    The callable object accepts a single parameter: the object to filter
    and returns a boolean if the object matches the filters or not.
    """

    def __init__(
        self,
        function_map: Optional[Dict[str, Callable]] = None,
        attribute_map: Optional[Dict[str, str]] = None,
        use_getattr: bool = True,
        allow_nested_attributes: bool = True,
    ):
        """Constructs a NativeEvaluator.

        Args:
            function_map: a mapping of a function name to a callable
                function.
            attribute_map: a mapping of an external name to an internal
                field of the item to be filtered. The internal field
                specifier can be a JSON-Path that will be resolved against
                the passed in item.
        """
        self.function_map = function_map if function_map is not None else {}
        self.attribute_map = attribute_map
        self.use_getattr = use_getattr
        self.allow_nested_attributes = allow_nested_attributes
        self.locals: Dict[str, Any] = {}
        self.local_count = 0

    def _add_local(self, value: Any) -> str:
        "Add a value as a local variable to the expression."
        self.local_count += 1
        key = f"local_{self.local_count}"
        self.locals[key] = value
        return key

    def _resolve_attribute(self, name):
        """Helper to resolve an attribute, either directly or via the
        integrated ``attribute_map``
        """
        if self.attribute_map is not None:
            if name in self.attribute_map:
                path = self.attribute_map[name]
            elif "*" in self.attribute_map:
                path = self.attribute_map["*"].replace("*", name)
            allow_nested_attributes = True
        else:
            path = name
            allow_nested_attributes = self.allow_nested_attributes

        parts = path.split(".")
        if not allow_nested_attributes and len(parts) > 1:
            raise Exception("Nested attributes are not allowed")

        return parts

    @handle(ast.Not)
    def not_(self, node, sub):
        return f"(not {sub})"

    @handle(ast.And)
    def and_(self, node, lhs, rhs):
        return f"({lhs} and {rhs})"

    @handle(ast.Or)
    def or_(self, node, lhs, rhs):
        return f"({lhs} or {rhs})"

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        op = COMPARISON_MAP[node.op]
        return f"({lhs} {op} {rhs})"

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        if node.not_:
            return f"({low} > {lhs} or {lhs} > {high})"
        else:
            return f"({low} <= {lhs} <= {high})"

    @handle(ast.Like)
    def like(self, node, lhs):
        maybe_not_inv = "" if node.not_ else "not "
        regex = like_pattern_to_re(
            node.pattern, node.nocase, node.wildcard, node.singlechar, node.escapechar
        )
        key = self._add_local(regex)
        return f"({key}.match({lhs}) is {maybe_not_inv}None)"

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        maybe_not = "not" if node.not_ else ""
        opts = ", ".join([f"{opt}" for opt in options])
        return f"({lhs} {maybe_not} in ({opts}))"

    @handle(ast.IsNull)
    def null(self, node, lhs):
        maybe_not = "not " if node.not_ else ""
        return f"({lhs} is {maybe_not}None)"

    @handle(ast.Exists)
    def exists(self, node, lhs):
        parts = self._resolve_attribute(node.lhs.name)
        maybe_not = "not " if node.not_ else ""
        if self.use_getattr:
            cur = "item"
            for part in parts[:-1]:
                cur = f"getattr({cur}, {part!r}, None)"
            return f"({maybe_not}hasattr({cur}, {parts[-1]!r}))"
        else:
            getters = "".join(f".get({part!r}, {{}})" for part in parts[:-1])
            return f"{parts[-1]!r} {maybe_not}in item{getters}"

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node, lhs, rhs):
        return (
            f"(relate_intervals(to_interval({lhs}),"
            f"to_interval({rhs})) == "
            f"ast.TemporalComparisonOp.{node.op.name})"
        )

    @handle(ast.ArrayPredicate, subclasses=True)
    def array(self, node, lhs, rhs):
        op = ARRAY_COMPARISON_OP_MAP[node.op]
        return f"bool(set({lhs}) {op} set({rhs}))"

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        return f"(getattr(ensure_spatial({lhs}), " f"{node.op.value.lower()!r})({rhs}))"

    @handle(ast.Relate)
    def spatial_pattern(self, node, lhs, rhs):
        return f"(ensure_spatial({lhs}).relate_pattern({rhs}, {node.pattern!r}))"

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        bbox_local = self._add_local(
            shapely.geometry.Polygon.from_bounds(
                node.minx, node.miny, node.maxx, node.maxy
            )
        )
        return f"(ensure_spatial({lhs}).intersects({bbox_local}))"

    @handle(ast.Attribute)
    def attribute(self, node):
        parts = self._resolve_attribute(node.name)
        if self.use_getattr:
            cur = "item"
            for part in parts:
                cur = f"getattr({cur}, {part!r}, None)"
            return cur
        else:
            getters = "".join(f".get({part!r})" for part in parts)
            return f"item{getters}"

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node, lhs, rhs):
        op = ARITHMETIC_MAP[node.op]
        return f"({lhs}) {op} ({rhs})"

    @handle(ast.Function)
    def function(self, node, *arguments):
        args = ", ".join([f"({arg})" for arg in arguments])
        return f"{node.name}({args})"

    @handle(*values.LITERALS)
    def literal(self, node):
        key = self._add_local(node)
        return key

    @handle(values.Interval)
    def interval(self, node, low, high):
        return f"values.Interval({low}, {high})"

    @handle(values.Geometry)
    def geometry(self, node):
        key = self._add_local(shapely.geometry.shape(node))
        return key

    @handle(values.Envelope)
    def envelope(self, node):
        key = self._add_local(
            shapely.geometry.Polygon.from_bounds(node.x1, node.y1, node.x2, node.y2)
        )
        return key

    def adopt_result(self, result):
        """Turns the compiled expression into a callable object using
        ``eval``. Literals are passed in as well as the function map.
        """
        expression = f"lambda item: {result}"
        globals_ = {
            "relate_intervals": relate_intervals,
            "to_interval": to_interval,
            "ensure_spatial": ensure_spatial,
            "ast": ast,
            "values": values,
        }
        if not set(globals_).isdisjoint(set(self.function_map)):
            raise ValueError(
                f"globals collision {list(globals_)} and " f"{list(self.function_map)}"
            )

        globals_.update(self.function_map)
        globals_.update(self.locals)

        # clear any locals for later use
        self.locals.clear()

        return eval(expression, globals_)


MaybeInterval = Union[values.Interval, date, datetime, str, None]
InternalInterval = Tuple[Optional[datetime], Optional[datetime]]


def _interval_to_internal_interval(value: values.Interval) -> InternalInterval:
    low = value.start
    high = value.end

    # convert low and high dates to their respective datetime
    # by using 00:00 time for the low part and 23:59:59 for the high
    # part
    if isinstance(low, date):
        low = datetime.combine(low, time.min, timezone.utc)
    if isinstance(high, date):
        high = datetime.combine(high, time.max, timezone.utc)

    # low and high are now either datetimes, timedeltas or None

    if isinstance(low, timedelta):
        if isinstance(high, datetime):
            low = high - low
        else:
            raise ValueError(f"Cannot combine {low} with {high}")
    elif isinstance(high, timedelta):
        if isinstance(low, datetime):
            high = low + high
        else:
            raise ValueError(f"Cannot combine {low} with {high}")

    return (low, high)


def to_interval(value: MaybeInterval) -> InternalInterval:
    """Converts the given value to an interval tuple of ``start``/``stop``
    as Python datetime objects.

    - ``values.Interval`` objects are expanded to two datetimes:
        - two datetimes are returned as such
        - a date is transformed to a datetime, where the ``time``
          component is either ``time.min`` for start or ``time.max``
          for then end component.
        - if either the start or end is a ``timedelta`` object, that value
          is either added to the start value or subtracted from the end
          value.
    - ``date`` objects are transformed to two datetimes for the
      ``time.min`` and ``time.end`` of that date in UTC.
    - ``datetime`` and ``str`` objects are an interval with both
      start and end of the same value. Strings are parsed beforehand.
    - ``None`` is simply returned as ``(None, None)``
    """

    if isinstance(value, str):
        value = parse_datetime(value)
        if not value.tzinfo:
            value = value.replace(tzinfo=timezone.utc)
        return (value, value)

    elif isinstance(value, values.Interval):
        return _interval_to_internal_interval(value)

    elif isinstance(value, datetime):
        return (value, value)

    elif isinstance(value, date):
        return (
            datetime.combine(value, time.min, timezone.utc),
            datetime.combine(value, time.max, timezone.utc),
        )

    elif value is None:
        return (None, None)

    raise ValueError(f"Invalid type {type(value)}")


def relate_intervals(  # noqa: C901
    lhs: InternalInterval, rhs: InternalInterval
) -> ast.TemporalComparisonOp:
    """Relates two intervals (tuples of two ``datetime`` or ``None`` values)
    and returns the associated ``ast.TemporalComparisonOp`` value.
    """
    ll, lh = lhs
    rl, rh = rhs
    if ll is None or lh is None or rl is None or rh is None:
        # TODO: handle open ended intervals (None on either side)
        return ast.TemporalComparisonOp.DISJOINT
    elif lh < rl:
        return ast.TemporalComparisonOp.BEFORE
    elif ll > rh:
        return ast.TemporalComparisonOp.AFTER
    elif lh == rl:
        return ast.TemporalComparisonOp.MEETS
    elif ll == rh:
        return ast.TemporalComparisonOp.METBY
    elif ll < rl and rl < lh < rh:
        return ast.TemporalComparisonOp.TOVERLAPS
    elif rl < ll < rh and lh > rh:
        return ast.TemporalComparisonOp.OVERLAPPEDBY
    elif ll == rl and lh < rh:
        return ast.TemporalComparisonOp.BEGINS
    elif ll == rl and lh > rh:
        return ast.TemporalComparisonOp.BEGUNBY
    elif ll > rl and lh < rh:
        return ast.TemporalComparisonOp.DURING
    elif ll < rl and lh > rh:
        return ast.TemporalComparisonOp.TCONTAINS
    elif ll > rl and lh == rh:
        return ast.TemporalComparisonOp.ENDS
    elif ll < rl and lh == rh:
        return ast.TemporalComparisonOp.ENDEDBY
    elif ll == rl and lh == rh:
        return ast.TemporalComparisonOp.TEQUALS

    raise ValueError(f"Error relating intervals [{ll}, {lh}] and [{rl}, {rh}]")


def ensure_spatial(value: Any) -> shapely.geometry.base.BaseGeometry:
    """Ensures that a given value is a shapely geometry. If it is already
    it is passed through, otherwise it is tried to be parsed via
    ``shapely.geometry.shape``.
    """
    if isinstance(value, shapely.geometry.base.BaseGeometry):
        return value
    return shapely.geometry.shape(value)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

"""
Parser implementation for JFE. Spec here:
https://github.com/tschaub/ogcapi-features/tree/json-array-expression/extensions/cql/jfe
"""

import json
from datetime import datetime
from typing import Any, Dict, List, Type, Union, cast

from ... import ast, values
from ...util import parse_datetime

COMPARISON_MAP: Dict[str, Type] = {
    "==": ast.Equal,
    "!=": ast.NotEqual,
    "<": ast.LessThan,
    "<=": ast.LessEqual,
    ">": ast.GreaterThan,
    ">=": ast.GreaterEqual,
}

SPATIAL_PREDICATES_MAP = {
    "intersects": ast.GeometryIntersects,
    "within": ast.GeometryWithin,
}

TEMPORAL_PREDICATES_MAP = {
    "before": ast.TimeBefore,
    "after": ast.TimeAfter,
    "during": ast.TimeDuring,
}

ARITHMETIC_MAP = {
    "+": ast.Add,
    "-": ast.Sub,
    "*": ast.Mul,
    "/": ast.Div,
}

FUNCTION_MAP = {
    "%": "mod",
    "^": "pow",
}

ParseResult = Union[
    ast.Node,
    str,
    float,
    int,
    datetime,
    values.Geometry,
    values.Interval,
    Dict[Any, Any],  # TODO: for like wildcards.
]


def _parse_node(node: Union[list, dict]) -> ParseResult:  # noqa: C901
    if isinstance(node, (str, float, int)):
        return node
    elif isinstance(node, dict):
        # wrap geometry, we say that the 'type' property defines if it is a
        # geometry
        if "type" in node:
            return values.Geometry(node)

        # just return objects for example `like` wildcards
        else:
            return node

    if not isinstance(node, list):
        raise ValueError(f"Invalid node class {type(node)}")

    op = node[0]
    arguments = [_parse_node(sub) for sub in node[1:]]

    if op in ["all", "any"]:
        cls = ast.And if op == "all" else ast.Or
        return cls.from_items(*arguments)

    elif op == "!":
        return ast.Not(*cast(List[ast.Node], arguments))

    elif op in COMPARISON_MAP:
        return COMPARISON_MAP[op](*arguments)

    elif op == "like":
        wildcard = "%"
        if len(arguments) > 2:
            wildcard = cast(dict, arguments[2]).get("wildCard", "%")
        return ast.Like(
            cast(ast.Node, arguments[0]),
            cast(str, arguments[1]),
            nocase=False,
            wildcard=wildcard,
            singlechar=".",
            escapechar="\\",
            not_=False,
        )

    elif op == "in":
        assert isinstance(arguments[0], ast.Node)
        return ast.In(
            cast(ast.Node, arguments[0]),
            cast(List[ast.AstType], arguments[1:]),
            not_=False,
        )

    elif op in SPATIAL_PREDICATES_MAP:
        return SPATIAL_PREDICATES_MAP[op](*cast(List[ast.SpatialAstType], arguments))

    elif op in TEMPORAL_PREDICATES_MAP:
        # parse strings to datetimes
        dt_args = [
            parse_datetime(arg) if isinstance(arg, str) else arg for arg in arguments
        ]
        if len(arguments) == 3:
            if isinstance(dt_args[0], datetime) and isinstance(dt_args[1], datetime):
                dt_args = [
                    values.Interval(dt_args[0], dt_args[1]),
                    dt_args[2],
                ]
            if isinstance(dt_args[1], datetime) and isinstance(dt_args[2], datetime):
                dt_args = [
                    dt_args[0],
                    values.Interval(dt_args[1], dt_args[2]),
                ]

        return TEMPORAL_PREDICATES_MAP[op](*cast(List[ast.TemporalAstType], dt_args))

    # special property getters
    elif op in ["id", "geometry"]:
        return ast.Attribute(op)

    # normal property getter
    elif op == "get":
        return ast.Attribute(arguments[0])

    elif op == "bbox":
        pass  # TODO

    elif op in ARITHMETIC_MAP:
        return ARITHMETIC_MAP[op](*cast(List[ast.ScalarAstType], arguments))

    elif op in ["%", "floor", "ceil", "abs", "^", "min", "max"]:
        return ast.Function(
            FUNCTION_MAP.get(op, op), cast(List[ast.AstType], arguments)
        )

    raise ValueError(f"Invalid expression operation '{op}'")


def parse(jfe: Union[str, list, dict]) -> ast.Node:
    """Parses the given JFE expression (either a string or an already
    parsed JSON) to an AST.
    If a string is passed, it will be parsed as JSON.

    https://github.com/tschaub/ogcapi-features/tree/json-array-expression/extensions/cql/jfe
    """
    if isinstance(jfe, str):
        root = json.loads(jfe)
    else:
        root = jfe

    return cast(ast.Node, _parse_node(root))



from ... import ast
from .base import FESBaseParser
from .util import Element, ParseInput, handle


class FES11Parser(FESBaseParser):
    namespace = "http://www.opengis.net/ogc"

    @handle("Add")
    def add(
        self, node: Element, lhs: ast.ScalarAstType, rhs: ast.ScalarAstType
    ) -> ast.Node:
        return ast.Add(lhs, rhs)

    @handle("Sub")
    def sub(
        self, node: Element, lhs: ast.ScalarAstType, rhs: ast.ScalarAstType
    ) -> ast.Node:
        return ast.Sub(lhs, rhs)

    @handle("Mul")
    def mul(
        self, node: Element, lhs: ast.ScalarAstType, rhs: ast.ScalarAstType
    ) -> ast.Node:
        return ast.Mul(lhs, rhs)

    @handle("Div")
    def div(
        self, node: Element, lhs: ast.ScalarAstType, rhs: ast.ScalarAstType
    ) -> ast.Node:
        return ast.Div(lhs, rhs)

    @handle("PropertyName")
    def property_name(self, node):
        return ast.Attribute(node.text)


def parse(input_: ParseInput) -> ast.Node:
    return FES11Parser().parse(input_)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2022 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

"""
Elasticsearch filter evaluator.

Uses elasticsearch-dsl package to create filter objects.
"""


# pylint: disable=E1130,C0103,W0223

from datetime import date, datetime
from typing import Dict, Optional, Union

from elasticsearch_dsl import Q
from packaging.version import Version

from ... import ast, values
from ..evaluator import Evaluator, handle
from .util import like_to_wildcard

VERSION_7_10_0 = Version("7.10.0")


COMPARISON_OP_MAP = {
    ast.ComparisonOp.LT: "lt",
    ast.ComparisonOp.LE: "lte",
    ast.ComparisonOp.GT: "gt",
    ast.ComparisonOp.GE: "gte",
}


ARITHMETIC_OP_MAP = {
    ast.ArithmeticOp.ADD: "+",
    ast.ArithmeticOp.SUB: "-",
    ast.ArithmeticOp.MUL: "*",
    ast.ArithmeticOp.DIV: "/",
}


class ElasticSearchDSLEvaluator(Evaluator):
    """A filter evaluator for Elasticsearch DSL."""

    def __init__(
        self,
        attribute_map: Optional[Dict[str, str]] = None,
        version: Optional[Version] = None,
    ):
        self.attribute_map = attribute_map
        self.version = version or Version("7.1.0")

    @handle(ast.Not)
    def not_(self, _, sub):
        """Inverts a filter object."""
        return ~sub

    @handle(ast.And)
    def and_(self, _, lhs, rhs):
        """Joins two filter objects with an `and` operator."""
        return lhs & rhs

    @handle(ast.Or)
    def or_(self, _, lhs, rhs):
        """Joins two filter objects with an `or` operator."""
        return lhs | rhs

    @handle(ast.Equal, ast.NotEqual)
    def equality(self, node, lhs, rhs):
        """Creates a match filter."""
        q = Q("match", **{lhs: rhs})
        if node.op == ast.ComparisonOp.NE:
            q = ~q
        return q

    @handle(ast.LessThan, ast.LessEqual, ast.GreaterThan, ast.GreaterEqual)
    def comparison(self, node, lhs, rhs):
        """Creates a `range` filter."""
        return Q("range", **{lhs: {COMPARISON_OP_MAP[node.op]: rhs}})

    @handle(ast.Between)
    def between(self, node: ast.Between, lhs, low, high):
        """Creates a `range` filter."""
        q = Q("range", **{lhs: {"gte": low, "lte": high}})
        if node.not_:
            q = ~q
        return q

    @handle(ast.Like)
    def like(self, node: ast.Like, lhs):
        """Transforms the provided LIKE pattern to an Elasticsearch wildcard
        pattern. Thus, this only works properly on "wildcard" fields.
        Ignores case-sensitivity when Elasticsearch version is below 7.10.0.
        """
        pattern = like_to_wildcard(
            node.pattern, node.wildcard, node.singlechar, node.escapechar
        )
        expr: Dict[str, Union[str, bool]] = {
            "value": pattern,
        }
        if self.version >= VERSION_7_10_0:
            expr["case_insensitive"] = node.nocase

        q = Q("wildcard", **{lhs: expr})
        if node.not_:
            q = ~q
        return q

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        """Creates a `terms` filter."""
        q = Q("terms", **{lhs: options})
        if node.not_:
            q = ~q
        return q

    @handle(ast.IsNull)
    def null(self, node: ast.IsNull, lhs):
        """Performs a null check, by using the `exists` query on the given
        field.
        """
        q = Q("exists", field=lhs)
        if not node.not_:
            q = ~q
        return q

    @handle(ast.Exists)
    def exists(self, node: ast.Exists, lhs):
        """Performs an existense check, by using the `exists` query on the
        given field
        """
        q = Q("exists", field=lhs)
        if node.not_:
            q = ~q
        return q

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node: ast.TemporalPredicate, lhs, rhs):
        """Creates a filter to match the given temporal predicate"""
        op = node.op
        if isinstance(rhs, (date, datetime)):
            low = high = rhs
        else:
            low, high = rhs

        query = "range"
        not_ = False
        predicate: Dict[str, Union[date, datetime, str]]
        if op == ast.TemporalComparisonOp.DISJOINT:
            not_ = True
            predicate = {"gte": low, "lte": high}
        elif op == ast.TemporalComparisonOp.AFTER:
            predicate = {"gt": high}
        elif op == ast.TemporalComparisonOp.BEFORE:
            predicate = {"lt": low}
        elif (
            op == ast.TemporalComparisonOp.TOVERLAPS
            or op == ast.TemporalComparisonOp.OVERLAPPEDBY
        ):
            predicate = {"gte": low, "lte": high}
        elif op == ast.TemporalComparisonOp.BEGINS:
            query = "term"
            predicate = {"value": low}
        elif op == ast.TemporalComparisonOp.BEGUNBY:
            query = "term"
            predicate = {"value": high}
        elif op == ast.TemporalComparisonOp.DURING:
            predicate = {"gt": low, "lt": high, "relation": "WITHIN"}
        elif op == ast.TemporalComparisonOp.TCONTAINS:
            predicate = {"gt": low, "lt": high, "relation": "CONTAINS"}
        # elif op == ast.TemporalComparisonOp.ENDS:
        #     pass
        # elif op == ast.TemporalComparisonOp.ENDEDBY:
        #     pass
        # elif op == ast.TemporalComparisonOp.TEQUALS:
        #     pass
        # elif op == ast.TemporalComparisonOp.BEFORE_OR_DURING:
        #     pass
        # elif op == ast.TemporalComparisonOp.DURING_OR_AFTER:
        #     pass
        else:
            raise NotImplementedError(f"Unsupported temporal operator: {op}")

        q = Q(
            query,
            **{lhs: predicate},
        )
        if not_:
            q = ~q
        return q

    @handle(
        ast.GeometryIntersects,
        ast.GeometryDisjoint,
        ast.GeometryWithin,
        ast.GeometryContains,
    )
    def spatial_comparison(self, node: ast.SpatialComparisonPredicate, lhs: str, rhs):
        """Creates a geo_shape query for the give spatial comparison
        predicate.
        """
        return Q(
            "geo_shape",
            **{
                lhs: {
                    "shape": rhs,
                    "relation": node.op.value.lower(),
                },
            },
        )

    @handle(ast.BBox)
    def bbox(self, node: ast.BBox, lhs):
        """Performs a geo_shape query for the given bounding box.
        Ignores CRS parameter, as it is not supported by Elasticsearch.
        """
        return Q(
            "geo_shape",
            **{
                lhs: {
                    "shape": self.envelope(
                        values.Envelope(node.minx, node.maxx, node.miny, node.maxy)
                    ),
                    "relation": "intersects",
                },
            },
        )

    @handle(ast.Attribute)
    def attribute(self, node: ast.Attribute):
        """Attribute mapping from filter fields to elasticsearch fields.
        If an attribute mapping is provided, it is used to look up the
        field name from there.
        """
        if self.attribute_map is not None:
            return self.attribute_map[node.name]
        return node.name

    # @handle(ast.Arithmetic, subclasses=True)
    # def arithmetic(self, node: ast.Arithmetic, lhs, rhs):
    #     op = ARITHMETIC_OP_MAP[node.op]
    #     return f"({lhs} {op} {rhs})"

    # @handle(ast.Function)
    # def function(self, node, *arguments):
    #     func = self.function_map[node.name]
    #     return f"{func}({','.join(arguments)})"

    @handle(*values.LITERALS)
    def literal(self, node):
        """Literal values are directly passed to elasticsearch-dsl"""
        return node

    @handle(values.Geometry)
    def geometry(self, node: values.Geometry):
        """Geometry values are converted to a GeoJSON object"""
        return node.geometry

    @handle(values.Envelope)
    def envelope(self, node: values.Envelope):
        """Envelope values are converted to an GeoJSON Elasticsearch
        extension object."""
        return {
            "type": "envelope",
            "coordinates": [
                [
                    min(node.x1, node.x2),
                    max(node.y1, node.y2),
                ],
                [
                    max(node.x1, node.x2),
                    min(node.y1, node.y2),
                ],
            ],
        }


def to_filter(
    root,
    attribute_map: Optional[Dict[str, str]] = None,
    version: Optional[str] = None,
):
    """Shorthand function to convert a pygeofilter AST to an Elasticsearch
    filter structure.
    """
    return ElasticSearchDSLEvaluator(
        attribute_map, Version(version) if version else None
    ).evaluate(root)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2025 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import logging
import sys

import click

from .parsers.cql_json.parser import parse as parse_cql_json
from .parsers.cql2_json.parser import parse as parse_cql2_json
from .parsers.cql2_text.parser import parse as parse_cql2_text
from .parsers.ecql.parser import parse as parse_ecql
from .parsers.fes.parser import parse as parse_fes
from .parsers.jfe.parser import parse as parse_jfe
from .version import __version__

__all__ = ["__version__"]

PARSERS = {
    'cql_json': parse_cql_json,
    'cql2_json': parse_cql2_json,
    'cql2_text': parse_cql2_text,
    'ecql': parse_ecql,
    'fes': parse_fes,
    'jfe': parse_jfe
}


def CLI_OPTION_VERBOSITY(f):
    """Setup click logging output"""
    def callback(ctx, param, value):
        if value is not None:
            logging.basicConfig(stream=sys.stdout,
                                level=getattr(logging, value))
        return True

    return click.option('--verbosity', '-v',
                        type=click.Choice(['ERROR', 'WARNING', 'INFO', 'DEBUG']),
                        help='Verbosity',
                        callback=callback)(f)


@click.group()
@click.version_option(version=__version__)
def cli():
    pass


@cli.command()
@click.pass_context
@click.argument('parser', type=click.Choice(PARSERS.keys()))
@click.argument('query')
@CLI_OPTION_VERBOSITY
def parse(ctx, parser, query, verbosity):
    """Parse a query into an abstract syntax tree"""

    click.echo(f'Parsing {parser} query into AST')
    try:
        click.echo(PARSERS[parser](query))
    except Exception as err:
        raise click.ClickException(err)


cli.add_command(parse)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Magnar Martinsen <magnarem@met.no>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2025 Norwegian Meteorological Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

"""
Apache Solr filter evaluator.

Uses native Python to return dict of JSON request payload
"""

# pylint: disable=E1130,C0103,W0223
from datetime import date, datetime
from typing import Optional

import shapely.wkt
from packaging.version import Version
from pygeoif import shape

from ... import ast, values
from ..evaluator import Evaluator, handle
from .util import like_to_wildcard

VERSION_9_8_1 = Version("9.8.1")

COMPARISON_OP_MAP = {
    ast.ComparisonOp.EQ: "{lhs}:{rhs}",
    ast.ComparisonOp.NE: "-{lhs}:{rhs}",
    ast.ComparisonOp.GT: "{lhs}:{{{rhs} TO *]",
    ast.ComparisonOp.GE: "{lhs}:[{rhs} TO *]",
    ast.ComparisonOp.LT: "{lhs}:[* TO {rhs}}}",
    ast.ComparisonOp.LE: "{lhs}:[* TO {rhs}]",
}

ARITHMETIC_OP_MAP = {
    ast.ArithmeticOp.ADD: "+",
    ast.ArithmeticOp.SUB: "-",
    ast.ArithmeticOp.MUL: "*",
    ast.ArithmeticOp.DIV: "/",
}


class SolrDSLQuery(dict):
    def __init__(self, query="*:*", filters=None):
        """
        Initialize a Solr JSON DSL query object.

        :param query: The main query (default is '*:*').
        :param filters: Optional filters to apply.
        """
        super().__init__()
        if isinstance(query, (str, dict)):
            self["query"] = query
        else:
            raise ValueError(f"Unsupported query type: {type(query)}")

        if filters is not None:
            if "filter" not in self:
                self["filter"] = []
            if isinstance(filters, str):
                self.add_filter(filters)
            if isinstance(filters, list):
                self["filter"] = filters

    def add_filter(self, filter_query):
        """
        Adds a filter query to the JSON DSL.

        :param filter_query: The filter query to add.
        """
        if "filter" not in self:
            self["filter"] = []
        self["filter"].append(filter_query)


class SOLRDSLEvaluator(Evaluator):
    """
    A filter evaluator for Apache Solr

    This evaluator uses the solr.SpatialRecursivePrefixTreeFieldType
    with the JTS context for querying on geometries, and the solr.DateRangeField
    for querying date ranges. See the test_evaluator py in this project
    for field definitions.
    """

    def __init__(
        self,
        attribute_map: Optional[dict[str, str]] = None,
        version: Optional[Version] = None,
    ):
        self.attribute_map = attribute_map
        self.version = version or Version("9.8.1")

    @handle(ast.And)
    def and_(self, _, lhs, rhs):
        """Joins two filter objects with an `and` operator."""
        # Extract the inner queries if lhs or rhs are SolrDSLQuery objects
        lhs = unwrap_query(lhs)
        rhs = unwrap_query(rhs)

        # Merge `must` and `must_not` properly
        combined_query = {"bool": {"must": []}}

        if "bool" in lhs and "must_not" in lhs["bool"]:
            # If `lhs` has a must_not clause, merge it
            combined_query["bool"]["must_not"] = lhs["bool"]["must_not"]
            lhs_must = {key: value for key, value in lhs["bool"].items() if key != "must_not"}
            if lhs_must:
                combined_query["bool"]["must"].append()
        else:
            # If `lhs` has no must_not clause, append it to must
            combined_query["bool"]["must"].append(lhs)

        if "bool" in rhs and "must_not" in rhs["bool"]:
            # If `rhs` has a must_not clause, merge it
            if "must_not" in combined_query["bool"]:
                combined_query["bool"]["must_not"].extend(rhs["bool"]["must_not"])
            else:
                combined_query["bool"]["must_not"] = rhs["bool"]["must_not"]
                rhs_must = {key: value for key, value in rhs["bool"].items() if key != "must_not"}
                if rhs_must:
                    combined_query["bool"]["must"].append()
        else:
            # If `rhs` has no must_not clause, append it to must
            combined_query["bool"]["must"].append(rhs)

        return SolrDSLQuery(combined_query)

    @handle(ast.Or)
    def or_(self, _, lhs, rhs):
        # Extract the inner queries if lhs or rhs are SolrDSLQuery objects
        lhs = unwrap_query(lhs)
        rhs = unwrap_query(rhs)

        # Merge `must` and `must_not` properly
        combined_query = {"bool": {"should": []}}

        if "bool" in lhs and "must_not" in lhs["bool"]:
            # If `lhs` has a must_not clause, merge it
            combined_query["bool"]["must_not"] = lhs["bool"]["must_not"]
            lhs_must = {key: value for key, value in lhs["bool"].items() if key != "must_not"}
            if lhs_must:
                combined_query["bool"]["should"].append()
        else:
            # If `lhs` has no must_not clause, append it to must
            combined_query["bool"]["should"].append(lhs)

        if "bool" in rhs and "must_not" in rhs["bool"]:
            # If `rhs` has a must_not clause, merge it
            if "must_not" in combined_query["bool"]:
                combined_query["bool"]["must_not"].extend(rhs["bool"]["must_not"])
            else:
                combined_query["bool"]["must_not"] = rhs["bool"]["must_not"]
                rhs_must = {key: value for key, value in rhs["bool"].items() if key != "must_not"}
                if rhs_must:
                    combined_query["bool"]["should"].append()
        else:
            # If `rhs` has no must_not clause, append it to must
            combined_query["bool"]["should"].append(rhs)

        return SolrDSLQuery(combined_query)

    @handle(ast.LessThan, ast.LessEqual, ast.GreaterThan, ast.GreaterEqual)
    def comparison(self, node, lhs, rhs):
        """
        Creates a range query for comparison operators.
        """
        return SolrDSLQuery(f"{COMPARISON_OP_MAP[node.op]}".format(lhs=lhs, rhs=rhs))

    @handle(ast.Between)
    def between(self, node: ast.Between, lhs, low, high):
        """
        Creates a range query for between conditions.
        """
        range_query = f"{lhs}:[{low} TO {high}]"
        if node.not_:
            # Negate the range query for NOT Between
            return SolrDSLQuery({"bool": {"must_not": [range_query]}})
        return SolrDSLQuery({"bool": {"must": [range_query]}})

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        """
        Creates a terms query for `IN` conditions.
        """
        options_str = " OR ".join(str(option) for option in options)
        terms_query = f"{lhs}:({options_str})"
        if node.not_:
            # Negate the terms query for NOT IN
            return SolrDSLQuery({"bool": {"must_not": [terms_query]}})
        return SolrDSLQuery({"bool": {"must": [terms_query]}})

    @handle(ast.IsNull)
    def null(self, node: ast.IsNull, lhs):
        """
        Creates a query to check for null values.
        """
        exists_query = f"(*:* -{lhs}:*)"
        if node.not_:
            exists_query = f"{lhs}:*"
        return SolrDSLQuery(exists_query)

    @handle(ast.Exists)
    def exists(self, node: ast.Exists, lhs):
        """
        Creates a query to check if a field exists.
        """
        exists_query = f"{lhs}:[* TO *]"
        if node.not_:
            exists_query = f"-{lhs}:[* TO *]"
        return SolrDSLQuery(exists_query)

    @handle(ast.Attribute)
    def attribute(self, node: ast.Attribute):
        """Attribute mapping from filter fields to Solr fields.
        If an attribute mapping is provided, it is used to look up the
        field name from there.
        """
        if self.attribute_map is not None:
            return self.attribute_map.get(node.name, node.name)
        return node.name

    @handle(*values.LITERALS)
    def literal(self, node):
        """Literal values are directly passed to Solr"""
        return node

    @handle(ast.Not)
    def not_(self, _, sub):
        """Inverts a filter object."""
        # Extract the inner query if sub is a SolrDSLQuery
        sub_query = sub["query"] if isinstance(sub, SolrDSLQuery) else sub

        # Handle the case where the sub-query is already a "must_not"
        if isinstance(sub_query, dict) and "bool" in sub_query and "must_not" in sub_query["bool"]:
            # If the sub-query is already a must_not, remove the negation
            return SolrDSLQuery({"bool": {"must": sub_query["bool"]["must_not"]}})

        # Otherwise, create a new must_not clause
        return SolrDSLQuery({"bool": {"must_not": [sub_query]}})

    @handle(ast.Like)
    def like(self, node: ast.Like, lhs):
        """Transforms the provided LIKE pattern to a Solr wildcard
        pattern. This only works properly on fields that are not tokenized.
        """
        pattern = like_to_wildcard(node.pattern, node.wildcard, node.singlechar, node.escapechar)
        if "*" in pattern:
            p = pattern.split("*")
            if p[0] == "":
                q = f"{{!complexphrase}}{lhs}:*{p[1].strip()}"
                if node.not_:
                    q = f'{{!complexphrase}}-{lhs}:"*{p[1].strip()}"'
            elif p[1] == "":
                q = f'{{!complexphrase}}{lhs}:"{p[0].strip()}*"'
                if node.not_:
                    q = f"{{!complexphrase}}-{lhs}:{p[0].strip()}*"
            else:
                q = f'{{!complexphrase}}{lhs}:"{p[0].strip()}"*"{p[1].strip()}"'
        elif "?" in pattern:
            q = f'{{!complexphrase}}{lhs}:"{pattern}"'
            if node.not_:
                q = f'{{!complexphrase}}-{lhs}:"{pattern}"'

        else:
            q = f'{lhs}:"{pattern}"'
            if node.not_:
                q = f"-{q}"
        return SolrDSLQuery(q)

    @handle(values.Geometry)
    def geometry(self, node: values.Geometry):
        """Geometry values are converted to a Solr spatial query."""
        """Convert to wkt and make sure polygons are counter clockwise"""
        geom_wkt = shape(node).wkt
        geom = shapely.wkt.loads(geom_wkt)
        if geom.geom_type == "Polygon" or geom.geom_type == "MultiPolygon":
            geom = geom.reverse() if not geom.exterior.is_ccw else geom
        return geom.wkt

    @handle(ast.Equal, ast.NotEqual)
    def equality(self, node, lhs, rhs):
        """
        Creates a term query for equality or inequality conditions.
        """
        if node.op == ast.ComparisonOp.EQ:
            # Use a term query for equality
            return SolrDSLQuery(f"{lhs}:{rhs}")
        elif node.op == ast.ComparisonOp.NE:
            # Use a boolean must_not query for inequality
            return SolrDSLQuery(f"-{lhs}:{rhs}")

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node: ast.TemporalPredicate, lhs, rhs):
        """Creates a filter to match the given temporal predicate"""
        op = node.op
        if isinstance(rhs, (date, datetime)):
            low = high = rhs.strftime("%Y-%m-%dT%H:%M:%SZ")
        else:
            low, high = rhs[0].strftime("%Y-%m-%dT%H:%M:%SZ"), rhs[1].strftime("%Y-%m-%dT%H:%M:%SZ")

        query = None
        if op == ast.TemporalComparisonOp.DISJOINT:
            query = f"-{lhs}:[{low} TO {high}]"
        elif op == ast.TemporalComparisonOp.AFTER:
            query = f"{lhs}:{{{high} TO *]"
        elif op == ast.TemporalComparisonOp.BEFORE:
            query = f"{lhs}:[* TO {low}}}"
        elif op == ast.TemporalComparisonOp.TOVERLAPS or op == ast.TemporalComparisonOp.OVERLAPPEDBY:
            query = f"{lhs}:[{low} TO {high}]"
        elif op == ast.TemporalComparisonOp.BEGINS:
            query = f"{lhs}:{low}"
        elif op == ast.TemporalComparisonOp.BEGUNBY:
            query = f"{lhs}:{high}"
        elif op == ast.TemporalComparisonOp.DURING:
            query = f"{lhs}:{{{low} TO {high}}}"
        elif op == ast.TemporalComparisonOp.TCONTAINS:
            query = f"{lhs}:[{low} TO {high}]"
        # elif op == ast.TemporalComparisonOp.ENDS:
        #     pass
        # elif op == ast.TemporalComparisonOp.ENDEDBY:
        #     pass
        # elif op == ast.TemporalComparisonOp.TEQUALS:
        #     pass
        # elif op == ast.TemporalComparisonOp.BEFORE_OR_DURING:
        #     pass
        # elif op == ast.TemporalComparisonOp.DURING_OR_AFTER:
        #     pass
        else:
            raise NotImplementedError(f"Unsupported temporal operator: {op}")

        return SolrDSLQuery(query)

    @handle(ast.GeometryIntersects, ast.GeometryDisjoint, ast.GeometryWithin, ast.GeometryContains, ast.GeometryEquals)
    def spatial_comparison(self, node: ast.SpatialComparisonPredicate, lhs: str, rhs):
        """Creates a spatial query for the given spatial comparison
        predicate.
        """
        query = {}
        # Solr need capitalized first letter of operator
        op = node.op.value.lower().capitalize()
        if op == "Disjoint":
            geo_filter = f"{{!field f={lhs} v='Intersects({rhs})'}}"
            query = {"bool": {"must_not": [geo_filter]}}
            return SolrDSLQuery(query)

        query = f"{{!field f={lhs} v='{op}({rhs})'}}"
        return SolrDSLQuery(query)

    @handle(ast.BBox)
    def bbox(self, node: ast.BBox, lhs):
        """Performs a spatial query for the given bounding box.
        Ignores CRS parameter, as it is not supported by Solr.
        """
        bbox = self.envelope(values.Envelope(node.minx, node.maxx, node.miny, node.maxy))
        query = f"{{!field f={lhs} v='Intersects({bbox})'}}"
        return SolrDSLQuery(query)

    # @handle(ast.Arithmetic, subclasses=True)
    # def arithmetic(self, node: ast.Arithmetic, lhs, rhs):
    #     op = ARITHMETIC_OP_MAP[node.op]
    #     return f"({lhs} {op} {rhs})"

    # @handle(ast.Function)
    # def function(self, node, *arguments):
    #     func = self.function_map[node.name]
    #     return f"{func}({','.join(arguments)})"

    @handle(values.Envelope)
    def envelope(self, node: values.Envelope):
        """Envelope values are converted to an WKT ENVELOPE for Solr."""
        min_x = float(min(node.x1, node.x2))
        max_x = float(max(node.x1, node.x2))
        min_y = float(min(node.y1, node.y2))
        max_y = float(max(node.y1, node.y2))
        return f"ENVELOPE({min_x}, {max_x}, {max_y}, {min_y})"


def to_filter(
    root,
    attribute_map: Optional[dict[str, str]] = None,
    version: Optional[Version] = None,
):
    """Shorthand function to convert a pygeofilter AST to an Apache Solr
    filter structure.
    """
    return SOLRDSLEvaluator(attribute_map, Version(version) if version else None).evaluate(root)


def unwrap_query(obj):
    """Extract the inner query from a SolrDSLQuery or return the object directly."""
    if isinstance(obj, SolrDSLQuery):
        # Return the inner query only if it is not empty
        return obj.get("query", {})
    return obj
